# Copyright (c) 2017 Arista Networks, Inc.  All rights reserved.
# Arista Networks, Inc. Confidential and Proprietary.

import BasicCli
import BasicCliUtil
import CliParser
import Cell
import LazyMount
import Tac
import Tracing
import CliToken.Clear
from CliPlugin import Ssl
from CliPlugin import Pki
from CliPlugin import Security
from CertRotationModel import CertificateRotation
from CertRotationModel import CertificateRotations
from CertRotationModel import RotationGenerateSignRequestStatus
from CertRotationModel import RotationImportStatus
from CertRotationModel import RotationCommitStatus
from CertRotationModel import RotationClearStatus
from CertRotationModel import RotationCommitSslProfileStatus
from RotationMgr import RotationMgr
from SslCertKey import SslCertKeyError

__defaultTraceHandle__ = Tracing.Handle( 'CertRotation' )
t0 = Tracing.trace0
Constants = Tac.Type( "Mgmt::Security::Ssl::Constants" ) 

rotationMgr = None
sslConfig = None
redundancyStatus = None
rotationNotReadyError = "certificate rotation not ready"

#------------------------------------------------------------------------------------
# security pki certificate generate signing-request
#      rotation ssl profile <name> key generate rsa <2048|3072|4096>
#      [ import-timeout <minutes> ]
#      [ digest <sha256|sha384|sha512> ]
#      parameters common-name <common-name>
#                             [ country <country-code> ]
#                             [ state <state-name> ]
#                             [ locality <locality-name> ]
#                             [ organization <org-name> ]
#                             [ organization-unit <org-unit-name> ]
#                             [ email <email> ]
#                             [ subject-alternative-name [ ip <ip1 ip2 ..> ]
#                                                        [ dns <name1 name2 ..> ]
#                                                        [ email <em1 em2 ..> ] ] ]
#------------------------------------------------------------------------------------
tokenRotation = CliParser.KeywordRule( "rotation",
      helpdesc="Rotation command",
      guard=CliParser.standbyGuard )
tokenImportTimeout = CliParser.KeywordRule( "import-timeout",
      helpdesc=( "Certificate import timeout in minutes after "
                 "which rotation expires" ) )
minutesRule = CliParser.RangeRule(
      Constants.minImportTimeoutMins, Constants.maxImportTimeoutMins,
      helpdesc="Minutes" )
importRule = CliParser.OptionalRule( ( tokenImportTimeout, minutesRule ),
                                     ifNotValue=Constants.defImportTimeoutMins,
                                     name="importTimeout" )

def rotationGenerateCsr( mode, profileName, newKeyBits, importTimeout,
                         digest, signReqParams ):
   if not rotationMgr.isReady():
      mode.addError( rotationNotReadyError )
      return None

   try:
      signReqParams = Pki.getSignRequestParams( mode, signReqParams )
   except SslCertKeyError as e:
      mode.addError( "Error generating CSR (%s)" % ( str( e ) ) )
      return

   return rotationMgr.generateCsr( profileName, importTimeout, 
                                   digest, signReqParams,
                                   newKeyBits=newKeyBits )

BasicCli.registerShowCommand( Security.tokenSecurity,
                              Pki.tokenEnPki,
                              Pki.tokenCert,
                              Pki.tokenGenerate,
                              Pki.tokenSignRequest,
                              ( ">>profileName", tokenRotation, Ssl.tokenSsl,
                                Ssl.tokenProfile, Ssl.profileNameRule ),
                              ( ">>newKeyBits", Pki.tokenKeyPair,
                                Pki.tokenGenerate, Pki.algoRule ),
                              importRule,
                              Pki.digestRule,
                              Pki.paramRule,
                              rotationGenerateCsr,
                              noShow=True,
                              noMore=True,
                              privileged=True,
                              cliModel=RotationGenerateSignRequestStatus )

#------------------------------------------------------------------------------------
# security pki certificate rotation import <rotation-id>
#------------------------------------------------------------------------------------
tokenImport = CliParser.KeywordRule( "import",
      helpdesc="Import certificate for rotation ID" )
rotationIdRule = CliParser.PatternRule( '[a-f0-9]+', 
      helpname="UUID",
      helpdesc="Unique rotation ID" )

def rotationImportCertificate( mode, rotationId ):
   if not rotationMgr.isReady():
      mode.addError( rotationNotReadyError )
      return None
   
   cmd = "%s %s %s %s %s UUID" % ( Security.tokenSecurity, Pki.tokenEnPki,
                                   Pki.tokenCert, tokenRotation, tokenImport )
   certPem = BasicCliUtil.getMultiLineInput( mode, cmd,
                                             prompt="Enter TEXT certificate" )
   return rotationMgr.importCertificate( rotationId, certPem )
   
BasicCli.registerShowCommand( Security.tokenSecurity,
                              Pki.tokenEnPki,
                              Pki.tokenCert,
                              tokenRotation,
                              tokenImport,
                              ( ">>rotationId", rotationIdRule ),
                              rotationImportCertificate,
                              noShow=True,
                              noMore=True,
                              privileged=True,
                              cliModel=RotationImportStatus )
                              
#------------------------------------------------------------------------------------
# security pki certificate rotation commit <rotation-id>
#------------------------------------------------------------------------------------
tokenCommit = CliParser.KeywordRule( "commit",
      helpdesc="Commit certificate and key for rotation" )

def rotationCommit( mode, rotationId ):
   if not rotationMgr.isReady():
      mode.addError( rotationNotReadyError )
      return None
   
   return rotationMgr.commit( rotationId )

BasicCli.registerShowCommand( Security.tokenSecurity,
                              Pki.tokenEnPki,
                              Pki.tokenCert,
                              tokenRotation,
                              tokenCommit,
                              ( ">>rotationId", rotationIdRule ),
                              rotationCommit,
                              noShow=True,
                              noMore=True,
                              privileged=True,
                              cliModel=RotationCommitStatus )

#------------------------------------------------------------------------------------
# security pki certificate rotation commit ssl profile <profile-name> [dont-wait]
#          [ignore validity]
#------------------------------------------------------------------------------------
tokenDontWait = CliParser.KeywordRule( "dont-wait",
      helpdesc="Do not wait for ssl services to load new certificate" )
tokenIgnore = CliParser.KeywordRule( "ignore",
      helpdesc="Ignore checks during commit" )

def rotationCommitSslProfile( mode, profileName, dontWait, ignoreValidity ):
   if not rotationMgr.isReady():
      mode.addError( rotationNotReadyError )
      return None
   
   dontWait = bool( dontWait )
   validateDates = not bool( ignoreValidity )
   cmd = "%s %s %s %s %s %s %s WORD" % ( Security.tokenSecurity, Pki.tokenEnPki,
                                         Pki.tokenCert, tokenRotation, tokenCommit,
                                         Ssl.tokenSsl, Ssl.tokenProfile )
   key = BasicCliUtil.getMultiLineInput( mode, cmd,
                                         prompt="Enter TEXT private key" )
   cert = BasicCliUtil.getMultiLineInput( mode, cmd,
                                          prompt="Enter TEXT certificate" )
   return rotationMgr.commitSslProfile( profileName, key, cert, validateDates,
                                        dontWait )
      
BasicCli.registerShowCommand( Security.tokenSecurity,
                              Pki.tokenEnPki,
                              Pki.tokenCert,
                              tokenRotation,
                              tokenCommit,
                              ( ">>profileName", Ssl.tokenSsl, 
                                Ssl.tokenProfile, Ssl.profileNameRule ),
                              [ ">>dontWait", tokenDontWait ],
                              [ ">>ignoreValidity", ( tokenIgnore, 
                                                      Pki.tokenValidity ) ],
                              rotationCommitSslProfile,
                              noShow=True,
                              noMore=True,
                              privileged=True,
                              cliModel=RotationCommitSslProfileStatus )

#------------------------------------------------------------------------------------
# clear security pki certificate rotation <rotation-id>
#------------------------------------------------------------------------------------
def rotationClear( mode, rotationId ):
   if not rotationMgr.isReady():
      mode.addError( rotationNotReadyError )
      return None
   
   return rotationMgr.clear( rotationId )

BasicCli.registerShowCommand( CliToken.Clear.clear,
                              Security.tokenSecurity,
                              Pki.tokenEnPki,
                              Pki.tokenCert,
                              tokenRotation,
                              ( ">>rotationId", rotationIdRule ),
                              rotationClear,
                              noShow=True,
                              noMore=True,
                              privileged=True,
                              cliModel=RotationClearStatus )

#------------------------------------------------------------------------------------
# show security pki certificate rotation [ detail ]
#------------------------------------------------------------------------------------
class ShowPkiCertRotations( object ):
   syntax = '''show security pki certificate rotation [ detail ]'''
   data = { 'security': Security.tokenSecurity,
            'pki': Pki.tokenEnPki,
            'certificate': Pki.tokenCert,
            'rotation': tokenRotation,
            'detail': ' Show additional rotation information' }

   def handler( self, mode, args ):
      if not rotationMgr.isReady():
         mode.addError( rotationNotReadyError )
         return None
      detail = 'detail' in args
      result = CertificateRotations( _detail=detail )
      for rotationId, rr in rotationMgr.getAllRotationReqs().iteritems():
         rotationModel = rr.getCliRotationModel()
         if rotationModel is None:
            continue
         result.rotations[ rotationId ] = rotationModel
      return result

BasicCli.registerLegacyShowCommandClass( ShowPkiCertRotations,
                                   cliModel=CertificateRotations )

#------------------------------------------------------------------------------------
# show security pki certificate rotation <rotation-id>
#------------------------------------------------------------------------------------
class ShowPkiCertRotation( object ):
   syntax = '''show security pki certificate rotation <rotation-id>'''
   data = { 'security': Security.tokenSecurity,
            'pki': Pki.tokenEnPki,
            'certificate': Pki.tokenCert,
            'rotation': tokenRotation,
            '<rotation-id>': rotationIdRule }

   def handler( self, mode, args ):
      if not rotationMgr.isReady():
         mode.addError( rotationNotReadyError )
         return None

      rotationReq = rotationMgr.getRotationReq( args[ '<rotation-id>' ] )
      if rotationReq is None:
         mode.addError( 'Rotation %s not found' % args[ '<rotation-id>' ] )
         return None

      result = rotationReq.getCliRotationModel()
      if result is None:
         mode.addError( 'Rotation %s not found' % args[ '<rotation-id>' ] )
      return result

BasicCli.registerLegacyShowCommandClass( ShowPkiCertRotation,
                                 cliModel=CertificateRotation )

# Called by tests only
def getRotationMgr():
   return rotationMgr

def Plugin( entityManager ):
   global sslConfig
   global redundancyStatus
   
   sslConfig = LazyMount.mount( entityManager, "mgmt/security/ssl/config",
                                "Mgmt::Security::Ssl::Config", "r" )
   sslServiceStatus = LazyMount.mount( entityManager, 
                                       "mgmt/security/ssl/serviceStatus",
                                       "Tac::Dir", "ri" )
   mg = entityManager.mountGroup()
   redundancyStatus = mg.mount( Cell.path( "redundancy/status" ),
                                "Redundancy::RedundancyStatus", "r" )
   
   def _mountsComplete():
      global rotationMgr      
      rotationMgr = RotationMgr( Constants.rotationBaseDirPath(), 
                                 redundancyStatus, sslConfig,
                                 sslServiceStatus )
   mg.close( _mountsComplete )
   
