# Copyright (c) 2016 Arista Networks, Inc.  All rights reserved.
# Arista Networks, Inc. Confidential and Proprietary.

import BasicCli
import BasicCliUtil
import CliParser
from CliPlugin import Ssl
import CliMode.Pki as PkiProfileMode
import ConfigMount
import LazyMount
import CliPlugin.Security as Security
import Tac, Assert, Tracing
import os
import IpGenAddr
import Arnet
import re
from SslCertKey import generateRsaPrivateKey
from SslCertKey import SslCertKeyError
from SslCertKey import generateCertificate
from SslCertKey import isSslDirsCreated

config = None
status = None
execRequest = None

Constants = Tac.Type( "Mgmt::Security::Ssl::Constants" )
Digest = Tac.Type( "Mgmt::Security::Ssl::Digest" )

__defaultTraceHandle__ = Tracing.Handle( 'PkiCli' )
t0 = Tracing.trace0
pkiNotReadyError = "PKI not ready"

# The config command 'pki profile' is still hidden.
# We are only giving enable mode 'pki' commands.
# Hence two tokens 
tokenPki = CliParser.HiddenRule( CliParser.KeywordRule(
   'pki',
   helpdesc='Configure PKI related options' ) )
tokenEnPki = CliParser.KeywordRule( 'pki', 
   helpdesc='Configure PKI related options' )

class PkiProfileConfigMode( PkiProfileMode.PkiProfileMode,
                            Ssl.ProfileConfigModeBase ):
   name = "PKI profile configuration"
   modeRule = CliParser.OrRule()

   def __init__( self, parent, session, profileName ):
      PkiProfileMode.PkiProfileMode.__init__( self, profileName )
      Ssl.ProfileConfigModeBase.__init__( self, parent, session, profileName,
                                          "profileTypePki" )

def _gotoPkiProfileConfigMode( mode, profileName ):
   childMode = mode.childMode( PkiProfileConfigMode, profileName=profileName )
   if childMode.profileConfig_.profileType != "profileTypePki":
      mode.addError( "Not a PKI profile" )
      return
   mode.session_.gotoChildMode( childMode )

def _noPkiProfile( mode, profileName ):
   profileConfig =  config.profileConfig.get( profileName )
   if profileConfig and profileConfig.profileType != "profileTypePki":
      mode.addError( "Not a PKI profile" )
      return
   del config.profileConfig[ profileName ]


Security.SecurityConfigMode.addCommand( ( tokenPki, Ssl.tokenProfile,
                                          Ssl.profileNameRule,
                                          _gotoPkiProfileConfigMode ) )

Security.SecurityConfigMode.addCommand( ( BasicCli.noOrDefault,
                                          tokenPki, Ssl.tokenProfile,
                                          Ssl.profileNameRule, _noPkiProfile ) )

#--------------------------------------------------
# [no|default] certificate command
#--------------------------------------------------
PkiProfileConfigMode.addCommand( Ssl.certificateCmd )
PkiProfileConfigMode.addCommand( Ssl.noCertificateCmd )

#--------------------------------------------------
# [no|default] trust certificate commannd
#--------------------------------------------------
PkiProfileConfigMode.addCommand( Ssl.trustCertificateCmd )
PkiProfileConfigMode.addCommand( Ssl.noTrustCertificateCmd )

#--------------------------------------------------
# [no|default] chain certificate command
#--------------------------------------------------
PkiProfileConfigMode.addCommand( Ssl.chainCertificateCmd )
PkiProfileConfigMode.addCommand( Ssl.noChainCertificateCmd )

#--------------------------------------------------
# [no|default] crl command
#--------------------------------------------------
PkiProfileConfigMode.addCommand( Ssl.crlCmd )
PkiProfileConfigMode.addCommand( Ssl.noCrlCmd )

#--------------------------------------------------------
# [no|default] certificate requirement extended-key-usage
#--------------------------------------------------------
PkiProfileConfigMode.addCommand( Ssl.verifyExtendedParametersCmd )
PkiProfileConfigMode.addCommand( Ssl.noVerifyExtendedParametersCmd )

#----------------------------------------------------------------------------
# [no|default] [trust|chain] certificate requirement basic-constraint ca true
#----------------------------------------------------------------------------
PkiProfileConfigMode.addCommand( Ssl.verifyBasicConstraintTrustCmd )
PkiProfileConfigMode.addCommand( Ssl.noVerifyBasicConstraintTrustCmd )
PkiProfileConfigMode.addCommand( Ssl.verifyBasicConstraintChainCmd )
PkiProfileConfigMode.addCommand( Ssl.noVerifyBasicConstraintChainCmd )

#--------------------------------------------------
# [no|default] [trust|chain] certificate requirement include root-ca
#--------------------------------------------------

PkiProfileConfigMode.addCommand( Ssl.chainIncludeRootCACmd )
PkiProfileConfigMode.addCommand( Ssl.noChainIncludeRootCACmd )

#---------------------------------------------------
# [no|default] certificate policy expiry-date ignore
#---------------------------------------------------
PkiProfileConfigMode.addCommand( Ssl.ignoreExpiryDateCmd  )
PkiProfileConfigMode.addCommand( Ssl.noIgnoreExpiryDateCmd  )

tokenKeyPair = CliParser.KeywordRule( "key", helpdesc="modify keys used" )
tokenGenerate = CliParser.KeywordRule( 'generate', 
                                       helpdesc='create new item',
                                       guard=CliParser.standbyGuard )
tokenRSA = CliParser.KeywordRule( 'rsa', 
                                  helpdesc='Use RSA algorithm' )
validRsaKeySizes = [ 2048, 3072, 4096 ]
rsaKeySizeTokens = []
for bitLength in validRsaKeySizes:
   tokenRsaBitSize = CliParser.KeywordRule( '%d' % bitLength,
                                             helpdesc='Use %d-bit keys' % bitLength )
   rsaKeySizeTokens.append( tokenRsaBitSize )
tokenRsaKeySize = CliParser.OrRule( *rsaKeySizeTokens, name='rsaBitLength' )
rsaAlgo = CliParser.ConcatRule( tokenRSA, tokenRsaKeySize, )
algoRule = CliParser.OrRule( rsaAlgo, name='keyAlgo' )

#------------------------------------------------------------------------------------
# security pki key generate rsa <size> <label>
#------------------------------------------------------------------------------------

def pkiGenerateKey( mode, key, keyAlgo ):
   """
   Generate a key pair for use in PKI.

   Saves under sslkey: with key  if successful 
   """
   if not isSslDirsCreated():
      mode.addError( pkiNotReadyError )
      return
   
   genericError = "Error generating key"
   # Assume that key is always RSA for now
   # keyAlgo holds the key size
   Assert.assertIn( int( keyAlgo ), validRsaKeySizes )
   keyFile = os.path.join( Constants.keysDirPath(), key )
   try:
      generateRsaPrivateKey( keyFile, int( keyAlgo ) )
   except SslCertKeyError as e:
      mode.addError( "%s (%s)" % ( genericError, str( e ) ) )
   except StandardError as e:
      mode.addError( "%s" % genericError )

BasicCli.EnableMode.addCommand( ( Security.tokenSecurity,
                                  tokenEnPki, tokenKeyPair, tokenGenerate,
                                  algoRule, Ssl.keyNameRule, pkiGenerateKey ) )

#------------------------------------------------------------------------------------
# security pki certificate generate { signing-request | self-signed <cert-name> }
#             [ key <key-name> [ generate rsa <2048|3072|4096> ] ] 
#             [ digest <sha256|sha384|sha512> ]
#             [ validity <days> ]
#             [ parameters common-name <common-name>
#                          [ country <country-code> ]
#                          [ state <state-name> ]
#                          [ locality <locality-name> ]
#                          [ organization <org-name> ]
#                          [ organization-unit <org-unit-name> ]
#                          [ email <email> ]
#                          [ subject-alternative-name [ ip <ip1 ip2 ..> ] 
#                                                     [ dns <name1 name2 ..> ] 
#                                                     [ email <em1 em2 ..> ] ] ]
#------------------------------------------------------------------------------------

def passTheBuck( mode, **args ):
   return args

tokenCert = CliParser.KeywordRule( 'certificate',
      helpdesc='work with x509 certificate' )
tokenSignRequest = CliParser.KeywordRule( 'signing-request',
      helpdesc='Certificate Signing Request ( CSR )' )
tokenSelfSigned = CliParser.KeywordRule( 'self-signed',
      helpdesc="Self signed certificate" )

tokenDigest = CliParser.KeywordRule( 'digest', 
      helpdesc='Digest to sign with' )
tokenSha256 = CliParser.KeywordRule( Digest.sha256, 
      helpdesc='Use 256 bit SHA' )
tokenSha384 = CliParser.KeywordRule( Digest.sha384, 
      helpdesc='Use 384 bit SHA' )
tokenSha512 = CliParser.KeywordRule( Digest.sha512, 
      helpdesc='Use 512 bit SHA' )

tokenValidity = CliParser.KeywordRule( 'validity', 
      helpdesc='Validity of certificate' )
validityDaysRule = CliParser.RangeRule( 
      Constants.minCertValidity, Constants.maxCertValidity, 
      helpdesc="Days" )

tokenParameters = CliParser.KeywordRule( 'parameters',
      helpdesc="Signing request parameters" )
tokenCommonName = CliParser.KeywordRule( 'common-name',
      helpdesc="Common name for use in subject" )
tokenCountry = CliParser.KeywordRule( 'country',
      helpdesc="Two-Letter Country Code for use in subject" )
tokenState = CliParser.KeywordRule( 'state',
      helpdesc="State for use in subject" )
tokenLocality = CliParser.KeywordRule( 'locality',
      helpdesc="Locality Name for use in subject" )
tokenOrganization = CliParser.KeywordRule( 'organization',
      helpdesc="Organization Name for use in subject" )
tokenOrgUnit = CliParser.KeywordRule( 'organization-unit',
      helpdesc="Organization Unit Name for use in subject" )
tokenEmail = CliParser.KeywordRule( 'email',
      helpdesc="Email address for use in subject" )
quotedParamRule = CliParser.QuotedStringRule()
emailAddrRule = CliParser.PatternRule( r'\S+', 
      helpname="WORD",
      helpdesc='Email address' )

tokenSan = CliParser.KeywordRule( 'subject-alternative-name',
      helpdesc="Subject alternative name extension" )
tokenIp = CliParser.KeywordRule( 'ip',
      helpdesc="IP addresses for use in subject-alternative-name" )
ipAddrRule = CliParser.TokenRule(
      matcher=IpGenAddr.IpGenAddrMatcher( "IP address" ) )
ipIterRule = CliParser.IterationRule( ipAddrRule ) 

tokenDns = CliParser.KeywordRule( 'dns',
      helpdesc="DNS names for use in subject-alternative-name" )
dnsNameRe = r'[0-9a-zA-Z_\.-]+'
dnsNameRule = CliParser.PatternRule( 
      r'^(?!email$|ip$)(%s)' % dnsNameRe,
      helpname="WORD",
      helpdesc="DNS name" )
dnsIterRule = CliParser.IterationRule( dnsNameRule )

tokenSanEmail = CliParser.KeywordRule( 'email',
      helpdesc="Email addresses for use in subject-alternative-name" )
sanEmailAddrRule = CliParser.PatternRule( r'^(?!dns$|ip$)(\S+)',
      helpname="WORD",
      helpdesc='Email address' )
sanEmailIterRule = CliParser.IterationRule( sanEmailAddrRule )

sanParamsRule = CliParser.SetRule( 
                     ( ">>sanIp", tokenIp, ipIterRule ),
                     ( ">>sanDns", tokenDns, dnsIterRule ),
                     ( ">>sanEmailAddress", tokenSanEmail, sanEmailIterRule ),
                     minMembers=1 )

keyGenRule = CliParser.OptionalRule( ( tokenGenerate, algoRule ), 
                                     name="newKeyBits" )
keyRule = CliParser.OptionalRule( ( tokenKeyPair, Ssl.keyNameRule, 
                                    keyGenRule, passTheBuck ),
                                  name="keyParams" )

shaRule = CliParser.OrRule( tokenSha256, tokenSha384, tokenSha512 )
digestRule = CliParser.OptionalRule( ( tokenDigest, shaRule ), 
                                     ifNotValue=Constants.defaultDigest,
                                     name="digest" )

validityRule = CliParser.OptionalRule( ( tokenValidity, validityDaysRule ),
                                       ifNotValue=Constants.defaultCertValidity,
                                       name="validityDays" )

paramRule = CliParser.OptionalRule( 
   ( tokenParameters, 
     ( ">>commonName", tokenCommonName, quotedParamRule ),
     [ ">>country", ( tokenCountry, quotedParamRule ) ],
     [ ">>state", ( tokenState, quotedParamRule ) ],
     [ ">>locality", ( tokenLocality, quotedParamRule ) ],
     [ ">>orgName", ( tokenOrganization, quotedParamRule ) ],
     [ ">>orgUnitName", ( tokenOrgUnit, quotedParamRule ) ],
     [ ">>emailAddress", ( tokenEmail, emailAddrRule ) ],
     [ ">>san", ( tokenSan, sanParamsRule ) ], 
     passTheBuck ),
   name="signReqParams" )

# Taken from RFC5280, Upper Bounds ( page 123 of May 2008 rev. )
upperBounds = { 'country' : 2,
                'state' : 128,
                'locality' : 128,
                'orgName' : 64,
                'orgUnitName' : 64,
                'commonName' : 64,
                'emailAddress' : 128,
              }
 
def _validateParams( paramName, paramValue ):
   if not paramValue:
      return
   
   printName = { 'country' : 'Country code', 
                 'state' : 'State', 
                 'locality' : 'Locality', 
                 'orgName' : 'Organization name',
                 'orgUnitName' : 'Organization unit name', 
                 'commonName' : 'Common name', 
                 'emailAddress' : 'Email address' }
     
   if paramName in upperBounds:
      if len( paramValue ) > upperBounds[ paramName ]:
         raise SslCertKeyError( "%s can be at most %d characters." % ( 
            printName[ paramName ], upperBounds[ paramName ] ) )
     
   if paramName == "sanIp":
      for ip in paramValue:
         try:
            Arnet.IpGenAddr( ip )
         except ( IndexError, ValueError ):
            raise SslCertKeyError( "IP address '%s' is not a valid v4 or "
                                   "v6 address" % ip )
        
   if paramName == "sanDns":
      for dnsName in paramValue:
         if not re.match( "^%s$" % dnsNameRe, dnsName ):
            raise SslCertKeyError( "DNS name '%s' is not valid." )
           
   if paramName == "sanEmailAddress":
      for emailAddress in paramValue:
         if len( emailAddress ) > upperBounds[ 'emailAddress' ]:
            raise SslCertKeyError( "%s '%s' can be at most %d characters." % (
                printName[ 'emailAddress' ], emailAddress, 
                upperBounds[ 'emailAddress' ] ) )

  
def _getSignRequestParamsInteractive( mode ):
   """
   Takes in as single line inputs each attribute required for generating a
   Certificate Signing Request and checks for errors. If no errors are found,
   generate a CSR and output to terminal.
   """
   signReqParams = dict()
   
   inp = BasicCliUtil.getSingleLineInput( mode,
         "Common Name for use in subject: " )
   if not inp:
      raise SslCertKeyError( "Common Name is needed" )
   signReqParams[ "commonName" ] = inp if inp else None
   _validateParams( "commonName", signReqParams[ "commonName" ] )
   
   inp = BasicCliUtil.getSingleLineInput( mode,
         "Two-Letter Country Code for use in subject: " )
   signReqParams[ "country" ] = inp if inp else None
   _validateParams( "country", signReqParams[ "country" ] )
   
   inp = BasicCliUtil.getSingleLineInput( mode,
         "State for use in subject: " )
   signReqParams[ "state" ] = inp if inp else None
   _validateParams( "state", signReqParams[ "state" ] )
   
   inp = BasicCliUtil.getSingleLineInput( mode,
         "Locality Name for use in subject: " )
   signReqParams[ "locality" ] = inp if inp else None
   _validateParams( "locality", signReqParams[ "locality" ] )
   
   inp = BasicCliUtil.getSingleLineInput( mode,
         "Organization Name for use in subject: " )
   signReqParams[ "orgName" ] = inp if inp else None
   _validateParams( "orgName", signReqParams[ "orgName" ] )
   
   inp = BasicCliUtil.getSingleLineInput( mode,
         "Organization Unit Name for use in subject: " )
   signReqParams[ "orgUnitName" ] = inp if inp else None
   _validateParams( "orgUnitName", signReqParams[ "orgUnitName" ] )
   
   inp = BasicCliUtil.getSingleLineInput( mode,
         "Email address for use in subject: " )
   signReqParams[ "emailAddress" ] = inp if inp else None
   _validateParams( "emailAddress", signReqParams[ "emailAddress" ] )
  
   inp = BasicCliUtil.getSingleLineInput( mode,
      "IP addresses (space separated) for use in subject-alternative-name: " )
   signReqParams[ "sanIp" ] = inp.split() if inp else None
   _validateParams( "sanIp", signReqParams[ "sanIp" ] )
   
   inp = BasicCliUtil.getSingleLineInput( mode,
      "DNS names (space separated) for use in subject-alternative-name: " )
   signReqParams[ "sanDns" ] = inp.split() if inp else None
   _validateParams( "sanDns", signReqParams[ "sanDns" ] )
   
   inp = BasicCliUtil.getSingleLineInput( mode,
      "Email addresses (space separated) for use in subject-alternative-name: " )
   signReqParams[ "sanEmailAddress" ] = inp.split() if inp else None
   _validateParams( "sanEmailAddress", signReqParams[ "sanEmailAddress" ] )
   
   return signReqParams

def _getKeyParams( mode, keyParams ):
   key = None
   newKeyBits = None
     
   if not keyParams:
      key = BasicCliUtil.getSingleLineInput( mode, "PKI Key to use for CSR: " )
   else:
      key = keyParams[ "key" ]
      newKeyBits = keyParams[ "newKeyBits" ]
  
   if not key:
      raise SslCertKeyError( "Key is needed" )
   
   keyFilepath = os.path.join( Constants.keysDirPath(), key )

   if not newKeyBits and not os.path.isfile( keyFilepath ):
      raise SslCertKeyError( "Key not found under sslkey:" )
      
   return ( keyFilepath, newKeyBits )
  
def getSignRequestParams( mode, signReqParams ):
   if not signReqParams:
      signReqParams = _getSignRequestParamsInteractive( mode )
   else:
      sanParams = signReqParams.pop( "san" )
      if sanParams:
         # Create a dict and fill in missing keys with None
         sanParams = dict( sanParams )
         sanParams[ "sanIp" ] = sanParams.get( "sanIp", None )            
         sanParams[ "sanDns" ] = sanParams.get( "sanDns", None ) 
         sanParams[ "sanEmailAddress" ] = sanParams.get( "sanEmailAddress", None )
         if sanParams[ "sanIp" ]:
            sanParams[ "sanIp" ] = [ x.stringValue for x in sanParams[ "sanIp" ] ]
      else:
         sanParams = { "sanIp": None, "sanDns": None, "sanEmailAddress": None }
      
      signReqParams.update( sanParams )
   
      for p in [ "commonName", "country", "state", "locality",
                 "orgName", "orgUnitName", "emailAddress",
                 "sanIp", "sanDns", "sanEmailAddress" ]:
         _validateParams( p, signReqParams[ p ] )
   
   return signReqParams   
      
def pkiGenerateCsrOrCert( mode, keyParams, digest, signReqParams, 
                          cert=None, validityDays=None ):
   """
   Generate CSR or self signed certificate
   """

   if not isSslDirsCreated():
      mode.addError( pkiNotReadyError )
      return
   
   genericError = "Error generating %s" % ( "certificate" if cert else "CSR" )   
   
   try:
      ( keyFilepath, newKeyBits ) = _getKeyParams( mode, keyParams )
      genNewKey = True if newKeyBits else False

      signReqParams = getSignRequestParams( mode, signReqParams )
      certFilepath = os.path.join( Constants.certsDirPath(), cert ) if cert else None
      
      signRequest = False if cert else True
      ( csr, _ ) = generateCertificate( keyFilepath=keyFilepath,
                                        certFilepath=certFilepath, 
                                        signRequest=signRequest,
                                        genNewKey=genNewKey, 
                                        newKeyBits=newKeyBits,
                                        digest=digest,
                                        validity=validityDays,
                                        **signReqParams )
      if signRequest:
         mode.addMessage( csr )
      else:
         mode.addMessage( "certificate:%s generated" % cert )
   except SslCertKeyError as e:
      mode.addError( "%s (%s)" % ( genericError, str( e ) ) )
   except EnvironmentError as e:
      mode.addError( "%s (%s)" % ( genericError, e.strerror ) )
   except StandardError as e:
      mode.addError( "%s" % genericError )
      
BasicCli.EnableMode.addCommand( ( Security.tokenSecurity,
                                  tokenEnPki, tokenCert, tokenGenerate,
                                  tokenSignRequest,
                                  keyRule,
                                  digestRule,
                                  paramRule,
                                  pkiGenerateCsrOrCert ) )

BasicCli.EnableMode.addCommand( ( Security.tokenSecurity,
                                  tokenEnPki, tokenCert, tokenGenerate,
                                  ( ">>cert", tokenSelfSigned, 
                                    Ssl.certificateNameRule ),
                                  keyRule,
                                  digestRule,
                                  validityRule,
                                  paramRule,
                                  pkiGenerateCsrOrCert ) )

def Plugin( entityManager ):
   global config, status, execRequest
   config = ConfigMount.mount( entityManager, "mgmt/security/ssl/config",
                               "Mgmt::Security::Ssl::Config", "w" )
   status = LazyMount.mount( entityManager, "mgmt/security/ssl/status",
                             "Mgmt::Security::Ssl::Status", "r" )
   execRequest = LazyMount.mount( entityManager, "mgmt/security/ssl/execRequest",
                                  "Mgmt::Security::Ssl::ExecRequest", "w" )
