# Copyright (c) 2013 Arista Networks, Inc.  All rights reserved.
# Arista Networks, Inc. Confidential and Proprietary.

import array
import fcntl
import termios
import os

import BasicCli
import Cell
import CliMatcher
import CliModel
import CliParser
from CliPlugin import ConfigMgmtMode
import ConfigMount
import LazyMount
import re
import SecretCli
import Tac
from SysdbPlugin.Security import mgmtSecurityConfigType

# ------------------------------
# Register callback for the forwarding chip information

forwardingInfoCallbacks = []

def registerForwardingChipInfoCallback( callbackFunc ):
   forwardingInfoCallbacks.append( callbackFunc )

# ------------------------------
# Security config commands
# 
# From config mode
#    management security - enter security config mode
#

securityConfig = None
securityStatus = None

tokenSecurity = CliParser.KeywordRule(
   'security',
   helpdesc='Configure security related options' )

securityKwMatcher = CliMatcher.KeywordMatcher(
   'security',
   helpdesc='Configure security related options' )

class SecurityConfigMode( ConfigMgmtMode.ConfigMgmtMode ):

   name = "Security configuration"
   modeParseTree = CliParser.ModeParseTree()

   def __init__( self, parent, session ):
      ConfigMgmtMode.ConfigMgmtMode.__init__( self, parent, session, "security" )
      self.config_ = securityConfig

   def setEntropySource( self ):
      self.config_.entropySourceHardware = True

   def noEntropySource( self ):
      self.config_.entropySourceHardware = False

   def defaultEntropySource( self ):
      self.config_.entropySourceHardware = self.config_.defaultEntropySourceHardware

   def setEntropyServer( self ):
      self.config_.useEntropyServer = True

   def defaultEntropyServer( self ):
      self.config_.useEntropyServer = self.config_.defaultUseEntropyServer

   def setPasswordMinimumLength( self, length ):
      self.config_.minPasswordLength = length

   def noPasswordMinimumLength( self ):
      self.config_.minPasswordLength = 0

   def setPasswordEncryptionKeyCommon( self ):
      self.config_.commonKeyEnabled = True

   def noPasswordEncryptionKeyCommon( self ):
      self.config_.commonKeyEnabled = False

   def defaultPasswordEncryptionKeyCommon( self ):
      self.config_.commonKeyEnabled = self.config_.commonKeyEnabledDefault

def entropyHardwareSupported( mode, token ):
   if securityStatus.entropySourceHardwareSupported:
      return None
   return CliParser.guardNotThisPlatform

def gotoSecurityConfigMode( mode ):
   childMode = mode.childMode( SecurityConfigMode )
   mode.session_.gotoChildMode( childMode )

def noSecurityConfig( mode ):
   childMode = mode.childMode( SecurityConfigMode )
   childMode.noEntropySource()
   BasicCli.removeComment( childMode )
   import CliPlugin.Ssl
   import CliPlugin.SharedSecretProfileCli
   CliPlugin.Ssl.noSecurityConfig( mode )
   CliPlugin.SharedSecretProfileCli.noSecurityConfig( mode )


BasicCli.GlobalConfigMode.addCommand(
   ( ConfigMgmtMode.tokenManagement, tokenSecurity, gotoSecurityConfigMode ) )

BasicCli.GlobalConfigMode.addCommand(
   ( BasicCli.noOrDefault, ConfigMgmtMode.tokenManagement, tokenSecurity, 
     noSecurityConfig ) )

#------------------------------------------------------------------------------------
# The "entropy source" commands, in config mode.
#
# [no|default] entropy source hardware [exclusive]
#------------------------------------------------------------------------------------

tokenEntropy = CliParser.KeywordRule(
   'entropy',
   helpdesc='Entropy configuration' )

tokenSource = CliParser.KeywordRule(
   'source',
   helpdesc='Source of entropy' )

tokenHardware = CliParser.KeywordRule(
   'hardware',
   helpdesc='Set entropy source to hardware',
   guard=entropyHardwareSupported )

tokenNoHardware = CliParser.KeywordRule(
   'hardware',
   helpdesc='Set entropy source to hardware' )

tokenExclusive = CliParser.KeywordRule(
   'exclusive',
   helpdesc='Only use entropy from the hardware source' )

SecurityConfigMode.addCommand( ( tokenEntropy,
                                 tokenSource,
                                 tokenHardware,
                                 SecurityConfigMode.setEntropySource ) )
SecurityConfigMode.addCommand( ( BasicCli.tokenNo,
                                 tokenEntropy,
                                 tokenSource,
                                 tokenNoHardware,
                                 SecurityConfigMode.noEntropySource ) )
SecurityConfigMode.addCommand( ( BasicCli.tokenDefault,
                                 tokenEntropy,
                                 tokenSource,
                                 tokenNoHardware,
                                 SecurityConfigMode.defaultEntropySource ) )

# NOTE: This is a separate command since there are plans to make the set of entropy
# sources a set in the future
SecurityConfigMode.addCommand( ( tokenEntropy,
                                 tokenSource,
                                 tokenHardware,
                                 tokenExclusive,
                                 SecurityConfigMode.setEntropyServer ) )
SecurityConfigMode.addCommand( ( BasicCli.noOrDefault,
                                 tokenEntropy,
                                 tokenSource,
                                 tokenHardware,
                                 tokenExclusive,
                                 SecurityConfigMode.defaultEntropyServer ) )

#------------------------------------------------------------------------------------
# The "show management security" command
#
# show management security
#------------------------------------------------------------------------------------

cpuModelRegex = re.compile( "model name\s+: (.*)" )

tokenShowSecurity = CliParser.KeywordRule(
   'security',
   helpdesc="Show security status" )

def opt( var, default ):
   return var if var else default

class SecurityStatus( CliModel.Model ):
   cpuModel = CliModel.Str( help='CPU Model name and manufacturer', 
                              optional=True )
   switchChip = CliModel.Str( help='Model of the switch chip',
                                        optional=True )
   securityChipVersion = CliModel.Str( help='Security chip version. Not present on '
                                       'older platforms.', optional=True )
   hardwareEntropyEnabled = CliModel.Bool( help='True if the source of entropy is '
                                                'hardware' )
   entropyServerEnabled = CliModel.Bool( help='True if entropy server is serving '
                                       'entropy in the entropy queue' )
   hardwareEntropyQueueSize = CliModel.Int( help='Number of bytes of hardware '
                                                 'generated entropy in the queue',
                                            optional=True )

   def render( self ):
      hardwareEntropyMsg = 'enabled' if self.hardwareEntropyEnabled else 'disabled'
      if self.entropyServerEnabled:
         hardwareEntropyMsg += "\nEntropy Queue Size (bytes): %d" % \
                               self.hardwareEntropyQueueSize
      print 'CPU Model:             %s' % opt( self.cpuModel, 'Not available' )
      print 'Security Chip:         %s' % opt( self.securityChipVersion, 
                                                'Not available' )
      asicLine = 'Forwarding ASIC:      '
      print asicLine,
      if self.switchChip:
         lines = self.switchChip.splitlines()
         print lines[ 0 ]
         for remainingInfo in lines[ 1: ]:
            header = ' ' * ( len( asicLine ) + 1 )
            print header + remainingInfo
      else:
         print "Not available"
      print 'Hardware Entropy Generation is %s' % hardwareEntropyMsg
      print

def getBytesInFifoPipe( mgmtSecConfig ):
   try:
      fifoFd = os.open( mgmtSecConfig.entropyFifoPath,
                        os.O_RDONLY | os.O_NONBLOCK )
      bytesInPipe = array.array( 'i', [ 0 ] )
      assert fcntl.ioctl( fifoFd, termios.FIONREAD, bytesInPipe, 1 ) == 0
      os.close( fifoFd )
      return bytesInPipe[ 0 ]
   except OSError:
      return 0

def showSecurityStatus( mode, vrfName="" ):
   cpuInfo = open( '/proc/cpuinfo', 'r' ).read()
   cpuModelSearch = re.search( cpuModelRegex, cpuInfo )
   if cpuModelSearch:
      cpuModel = cpuModelSearch.group( 1 )
   else:
      cpuModel = None
   hardwareInfo = None
   for callback in forwardingInfoCallbacks:
      chipInfo = callback( mode )
      if chipInfo:
         hardwareInfo = chipInfo
         break

   res = SecurityStatus()
   res.cpuModel = opt( cpuModel, None )
   res.switchChip = opt( hardwareInfo, None )
   res.securityChipVersion = opt( securityStatus.securityChipVersion, None )
   res.hardwareEntropyEnabled = securityStatus.entropySourceHardwareEnabled
   res.entropyServerEnabled = securityStatus.entropyServerEnabled
   totalBytes = securityStatus.entropyQueueSizeBytes + \
                getBytesInFifoPipe( securityConfig )
   res.hardwareEntropyQueueSize = totalBytes
   
   return res

BasicCli.registerShowCommand( ConfigMgmtMode.tokenManagement,
                              tokenShowSecurity, showSecurityStatus, 
                              cliModel=SecurityStatus )

#------------------------------------------------------------------------------------
# The "password minimum length" commands, in config mode.
#
# [no|default] password minimum length <n>
#------------------------------------------------------------------------------------

tokenPassword = CliParser.KeywordRule(
   'password',
   helpdesc='Password configuration' )

tokenMinimum = CliParser.KeywordRule(
   'minimum',
   helpdesc='Minimum setting' )

tokenLength = CliParser.KeywordRule(
   'length',
   helpdesc='Number of characters' )

tokenLengthValue = CliParser.RangeRule( 1, 32, name='length', 
   helpdesc='Value between 1 and 32' )

SecurityConfigMode.addCommand( ( tokenPassword,
                                 tokenMinimum,
                                 tokenLength,
                                 tokenLengthValue,
                                 SecurityConfigMode.setPasswordMinimumLength ) )
SecurityConfigMode.addCommand( ( BasicCli.noOrDefault,
                                 tokenPassword,
                                 tokenMinimum,
                                 tokenLength,
                                 SecurityConfigMode.noPasswordMinimumLength ) )

#------------------------------------------------------------------------------------
# The "password encryption-key common" commands, in config mode.
#
# [no|default] password encryption-key common <n>
#------------------------------------------------------------------------------------
tokenEncryptionKey = CliParser.KeywordRule( 
   'encryption-key', 
   helpdesc='internal storage encryption-key')

tokenCommon = CliParser.KeywordRule( 
   'common', 
   helpdesc='use a common internal key')

SecurityConfigMode.addCommand( (
         tokenPassword,
         tokenEncryptionKey,
         tokenCommon,
         SecurityConfigMode.setPasswordEncryptionKeyCommon ) )

SecurityConfigMode.addCommand( (
         BasicCli.tokenNo,
         tokenPassword, 
         tokenEncryptionKey,
         tokenCommon,
         SecurityConfigMode.noPasswordEncryptionKeyCommon ) )

SecurityConfigMode.addCommand( (
         BasicCli.tokenDefault,
         tokenPassword, 
         tokenEncryptionKey,
         tokenCommon,
         SecurityConfigMode.defaultPasswordEncryptionKeyCommon ) )


def Plugin( entityManager ):
   global securityConfig, securityStatus
   securityConfig = ConfigMount.mount( entityManager, "mgmt/security/config",
                                       mgmtSecurityConfigType, "w" )
   securityStatus = LazyMount.mount( entityManager,
                                     Cell.path( "mgmt/security/status" ),
                                     "Mgmt::Security::Status", "r" )
   SecretCli.setMinPasswordLengthFunc( lambda: securityConfig.minPasswordLength )

