# Copyright (c) 2014 Arista Networks, Inc.  All rights reserved.
# Arista Networks, Inc. Confidential and Proprietary.

import BasicCli
import Tac
import CliParser
from CliRule import KeywordRule
import ConfigMgmtMode
import ConfigMount
import LazyMount
import CliPlugin.Security as Security
import CliMode.Ssl as SslProfileMode
import CliPlugin.SslModel as SslModel
import SslCliLib
import CliToken.Reset as CliReset
from SetRule import SetRule
import os
import os.path
import SslCertKey

config = None
status = None
execRequest = None
CertKeyPair = Tac.Type( "Mgmt::Security::Ssl::CertKeyPair" )
Constants = Tac.Type( "Mgmt::Security::Ssl::Constants" )

def _listdir( path ):
   if not os.path.isdir( path ):
      return []
   return os.listdir( path )

tokenSsl = CliParser.KeywordRule( 
   'ssl',
   helpdesc='Configure SSL related options' )
tokenProfile = CliParser.KeywordRule( 
   'profile',
   helpdesc='Configure SSL profile' )
tokenTls = CliParser.KeywordRule( 
   'tls',
   helpdesc='Configure TLS settings' )
tokenFips = CliParser.KeywordRule(
   'fips',
   helpdesc='Configure FIPS settings' )
tokenFipsRestrictions = CliParser.KeywordRule(
   'restrictions',
   helpdesc='Configure FIPS restrictions' )
tokenTlsVersions = CliParser.KeywordRule(
   'versions',
   helpdesc='Configure TLS versions' )
tokenTlsVersionsAdd = CliParser.KeywordRule(
   'add',
   helpdesc='Add versions to the current list' )
tokenTlsVersionsRemove = CliParser.KeywordRule(
   'remove',
   helpdesc='Remove versions from the current list' )
tokenCipherList = CliParser.KeywordRule(
   'cipher-list',
   helpdesc='Configure a cipher list' )
tokenCertificate = CliParser.KeywordRule( 
   'certificate',
   helpdesc='Configure certificate for self authentication' )
tokenKey = CliParser.KeywordRule( 
   'key',
   helpdesc='Configure key matching the certificate' )
tokenTrust = CliParser.KeywordRule( 
   'trust',
   helpdesc='Configure trusted certificate' )
tokenTrustCertificate = CliParser.KeywordRule( 
   'certificate',
   helpdesc='Certificate' )
tokenChain = CliParser.KeywordRule( 
   'chain',
   helpdesc='Configure chained certificate' )
tokenChainCertificate = CliParser.KeywordRule( 
   'certificate',
   helpdesc='Certificate' )
tokenCrl = CliParser.KeywordRule(
   'crl',
   helpdesc='Configure CRLs' )
tokenDh = CliParser.KeywordRule( 
   'diffie-hellman',
   helpdesc='diffie-hellman parameters' )
tokenEndEntity = CliParser.KeywordRule(
   'end-entity',
   helpdesc='Configure end-entiry certificate' )
tokenParameters = CliParser.KeywordRule(
   'parameters',
   helpdesc='diffie-hellman parameters' )
tokenRequirement = CliParser.KeywordRule(
   'requirement',
   helpdesc='Add a check to the certificate for validity' )
tokenRootCA = CliParser.KeywordRule(
   'root-ca',
   helpdesc='Root Certificate Authority' )
tokenPolicy = CliParser.KeywordRule(
   'policy',
   helpdesc='Policy parameters' )
tokenIgnore = CliParser.KeywordRule(
   'ignore',
   helpdesc='Ignore a requirement' )
tokenInclude = CliParser.KeywordRule(
   'include',
   helpdesc='Add a requirement to be included when validating' )
tokenExpiryDate = CliParser.KeywordRule(
   'expiry-date',
   helpdesc='Certificate expiry date' )
tokenExtended = CliParser.KeywordRule(
   'extended-key-usage',
   helpdesc='Certificate extended key usage extension' )
tokenBasicConstraint = CliParser.KeywordRule(
   'basic-constraint',
   helpdesc='Certificate basic constraint extension' )
tokenCa = CliParser.KeywordRule(
   'ca',
   helpdesc='Certificate authority attribute must satisfy the requirement' )
tokenTrue = CliParser.KeywordRule(
   'true',
   helpdesc='Attribute must be set to true' )
tokenVerify = CliParser.KeywordRule(
   'verify',
   helpdesc='verify' )

# Show tokens
tokenShowSsl = CliParser.KeywordRule( 
   'ssl',
   helpdesc='Show SSL status' )
tokenShowProfile = CliParser.KeywordRule( 
   'profile',
   helpdesc='Show SSL profile status' )
tokenShowCertificate = CliParser.KeywordRule(
   'certificate',
   helpdesc='Show certificate' )
tokenShowCrl = CliParser.KeywordRule(
   'crl',
   helpdesc='Show crl' )
tokenShowKey = CliParser.KeywordRule( 
   'key',
   helpdesc='Show key' )
tokenShowDh = CliParser.KeywordRule( 
   'diffie-hellman',
   helpdesc='Show diffie-hellman parameters in use' )

profileNameRule = CliParser.DynamicNameRule( lambda mode: config.profileConfig,
                                             'SSL profile name', 
                                             name='profileName' )
statusProfileNameRule = CliParser.DynamicNameRule( 
                                             lambda mode: status.profileStatus,
                                             'SSL profile name', 
                                             name='profileName' )
certificateNameRule = CliParser.DynamicNameRule( 
                                 lambda mode: _listdir( Constants.certsDirPath() ),
                                 'Certificate name', 
                                 name='certificate',
                                 pattern=r'[A-Za-z0-9_:{}\.\[\]-]+' )
keyNameRule = CliParser.DynamicNameRule( 
                                 lambda mode: _listdir( Constants.keysDirPath() ),
                                 'Key name', 
                                 name='key',
                                 pattern=r'[A-Za-z0-9_:{}\.\[\]-]+' )

def _getOutput( cmd, inp=None ):
   return Tac.run( cmd, asRoot=True, stdout=Tac.CAPTURE, stderr=Tac.CAPTURE,
         ignoreReturnCode=True, input=inp )

class ProfileConfigModeBase( BasicCli.ConfigModeBase ):

   def __init__( self, parent, session, profileName, profileType ):
      BasicCli.ConfigModeBase.__init__( self, parent, session )
      self.profileName = profileName

      if self.profileName not in config.profileConfig:
         self.profileConfig_ = config.newProfileConfig( self.profileName )
         self.profileConfig_.profileType = profileType
      else:
         self.profileConfig_ = config.profileConfig[ self.profileName ]
         if self.profileConfig_.profileType != profileType:
            return

   def enableTlsVersion( self, mask, add=False ):
      if add:
         mask |= self.profileConfig_.tlsVersion
      self.profileConfig_.tlsVersion = mask

   def disableTlsVersion( self, mask ):
      version = self.profileConfig_.tlsVersion & ~mask
      if not version:
         self.addError( "Cannot disable all TLS versions" )
         return
      self.profileConfig_.tlsVersion = version

   def enableFipsMode( self ):
      self.profileConfig_.fipsMode = True

   def disableFipsMode( self ):
      self.profileConfig_.fipsMode = False

   def enableCiphers( self, ciphers ):
      if 'Error' in _getOutput( [ 'openssl', 'ciphers', ciphers ] ):
         self.addError( 'Invalid cipher list.' )
         return
      self.profileConfig_.cipherSuite = ciphers

   def disableCiphers( self ):
      self.profileConfig_.cipherSuite = Constants.defaultCipherSuite()

   def setCertKey( self, certificate, key ):
      self.profileConfig_.certKeyPair = CertKeyPair( certificate, key )

   def noCertKey( self ):
      self.profileConfig_.certKeyPair = CertKeyPair( "", "" )

   def addTrustedCert( self, trustedCert ):
      if not trustedCert in self.profileConfig_.trustedCert:
         self.profileConfig_.trustedCert[ trustedCert ] = 1

   def noTrustedCert( self, trustedCert ):
      if trustedCert in self.profileConfig_.trustedCert:
         del self.profileConfig_.trustedCert[ trustedCert ]

   def addChainedCert( self, chainedCert ):
      if not chainedCert in self.profileConfig_.chainedCert:
         self.profileConfig_.chainedCert[ chainedCert ] = 1

   def noChainedCert( self, chainedCert ):
      if chainedCert in self.profileConfig_.chainedCert:
         del self.profileConfig_.chainedCert[ chainedCert ]

   def addCrl( self, crl ):
      if not crl in self.profileConfig_.crl:
         self.profileConfig_.crl[ crl ] = 1

   def noCrl( self, crl ):
      if crl in self.profileConfig_.crl:
         del self.profileConfig_.crl[ crl ]

   def enableExtendedParameters( self ):
      self.profileConfig_.verifyExtendedParameters = True

   def disableExtendedParameters( self ):
      self.profileConfig_.verifyExtendedParameters = False

   def enableVerifyBasicConstraintTrust( self ):
      self.profileConfig_.verifyBasicConstraintTrust = True

   def disableVerifyBasicConstraintTrust( self ):
      self.profileConfig_.verifyBasicConstraintTrust = False

   def enableVerifyBasicConstraintChain( self ):
      self.profileConfig_.verifyBasicConstraintChain = True

   def disableVerifyBasicConstraintChain( self ):
      self.profileConfig_.verifyBasicConstraintChain = False

   def enableVerifyExpiryDateEndCert( self ):
      self.profileConfig_.verifyExpiryDateEndCert = True

   def disableVerifyExpiryDateEndCert( self ):
      self.profileConfig_.verifyExpiryDateEndCert = False

   def setVerifyChainHasRootCA( self, setting ):
      self.profileConfig_.verifyChainHasRootCA = setting

class SslProfileConfigMode( SslProfileMode.SslProfileMode, 
                            ProfileConfigModeBase ):
   name = "SSL profile configuration"
   modeRule = CliParser.OrRule()

   def __init__( self, parent, session, profileName ):
      SslProfileMode.SslProfileMode.__init__( self, profileName )
      ProfileConfigModeBase.__init__( self, parent, session, profileName,
                                      "profileTypeSsl" )

def _gotoSslProfileConfigMode( mode, profileName ):
   childMode = mode.childMode( SslProfileConfigMode, profileName=profileName )
   if childMode.profileConfig_.profileType != "profileTypeSsl":
      mode.addError( "Not an SSL profile" )
      return
   mode.session_.gotoChildMode( childMode )

def _noSslProfile( mode, profileName ):
   profileConfig =  config.profileConfig.get( profileName )
   if profileConfig and profileConfig.profileType != "profileTypeSsl":
      mode.addError( "Not an SSL profile" )
      return
   del config.profileConfig[ profileName ]

def noSecurityConfig( mode ):
   for profileName in config.profileConfig:
      del config.profileConfig[ profileName ]
   
     
Security.SecurityConfigMode.addCommand( ( tokenSsl, tokenProfile,
                                          profileNameRule,
                                          _gotoSslProfileConfigMode) )

Security.SecurityConfigMode.addCommand( ( BasicCli.noOrDefault,
                                          tokenSsl, tokenProfile, profileNameRule, 
                                          _noSslProfile ) )

#--------------------------------------------------
# [no|default] certificate requirement extended-key-usage
#--------------------------------------------------
def _enableExtendedParameters( mode ):
   mode.enableExtendedParameters()

def _disableExtendedParameters( mode ):
   mode.disableExtendedParameters()


verifyExtendedParametersCmd = ( tokenCertificate, tokenRequirement, tokenExtended,
                                _enableExtendedParameters ) 

noVerifyExtendedParametersCmd = ( BasicCli.noOrDefault,
                                  tokenCertificate, tokenRequirement, tokenExtended,
                                  _disableExtendedParameters )

SslProfileConfigMode.addCommand( verifyExtendedParametersCmd )

SslProfileConfigMode.addCommand( noVerifyExtendedParametersCmd )

#--------------------------------------------------
# [no|default] [trust|chain] certificate requirement basic-constraints ca true
#--------------------------------------------------
def _enableVerifyBasicConstraintTrust( mode ):
   mode.enableVerifyBasicConstraintTrust()

def _disableVerifyBasicConstraintTrust( mode ):
   mode.disableVerifyBasicConstraintTrust()

def _enableVerifyBasicConstraintChain( mode ):
   mode.enableVerifyBasicConstraintChain()

def _disableVerifyBasicConstraintChain( mode ):
   mode.disableVerifyBasicConstraintChain()

verifyBasicConstraintTrustCmd = ( tokenTrust, tokenCertificate, tokenRequirement,
                                  tokenBasicConstraint, tokenCa, tokenTrue,
                                  _enableVerifyBasicConstraintTrust )

noVerifyBasicConstraintTrustCmd = ( BasicCli.noOrDefault,
                                    tokenTrust, tokenCertificate, tokenRequirement,
                                    tokenBasicConstraint, tokenCa, tokenTrue,
                                    _disableVerifyBasicConstraintTrust )

verifyBasicConstraintChainCmd = ( tokenChain, tokenCertificate, tokenRequirement,
                                  tokenBasicConstraint, tokenCa, tokenTrue,
                                  _enableVerifyBasicConstraintChain )

noVerifyBasicConstraintChainCmd = ( BasicCli.noOrDefault,
                                    tokenChain, tokenCertificate, tokenRequirement,
                                    tokenBasicConstraint, tokenCa, tokenTrue,
                                    _disableVerifyBasicConstraintChain )

SslProfileConfigMode.addCommand( verifyBasicConstraintTrustCmd )

SslProfileConfigMode.addCommand( noVerifyBasicConstraintTrustCmd )

SslProfileConfigMode.addCommand( verifyBasicConstraintChainCmd )

SslProfileConfigMode.addCommand( noVerifyBasicConstraintChainCmd )

#---------------------------------------------------
# [no|default] certificate policy expiry-date ignore
#---------------------------------------------------
def _enableIgnoreExpiryDate( mode ):
   mode.disableVerifyExpiryDateEndCert()

def _disableIgnoreExpiryDate( mode ):
   mode.enableVerifyExpiryDateEndCert()

ignoreExpiryDateCmd = ( tokenCertificate, tokenPolicy, tokenExpiryDate,
                        tokenIgnore, _enableIgnoreExpiryDate )

noIgnoreExpiryDateCmd = ( BasicCli.noOrDefault,
                          tokenCertificate, tokenPolicy, tokenExpiryDate,
                          tokenIgnore, _disableIgnoreExpiryDate )

SslProfileConfigMode.addCommand( ignoreExpiryDateCmd )

SslProfileConfigMode.addCommand( noIgnoreExpiryDateCmd )

#---------------------------------------------------
# [no|default] tls versions
# tls versions [add | remove] { 1.0 | 1.1 | 1.2 }...
#---------------------------------------------------
tlsVersionsMap = { '1.0': Constants.tlsv1,
                   '1.1': Constants.tlsv1_1,
                   '1.2': Constants.tlsv1_2 }
allTlsVersions = ( 'all', Constants.allTlsVersion )
tlsVersionsRule = SetRule( *sorted( KeywordRule( v, 'version {}'.format( v ),
                                                 value=lambda mode, match:
                                                       tlsVersionsMap[ match ] )
                                    for v in tlsVersionsMap ),
                           name='versions',
                           minMembers=1 )

def _enableTlsVersion( mode, versions=( allTlsVersions, ), add=False ):
   mode.enableTlsVersion( sum( v for __, v in versions ), add )

def _disableTlsVersion( mode, versions ):
   mode.disableTlsVersion( sum( v for __, v in versions ) )

SslProfileConfigMode.addCommand( ( BasicCli.noOrDefault, tokenTls,
                                   tokenTlsVersions, BasicCli.trailingGarbage,
                                   _enableTlsVersion ) )
SslProfileConfigMode.addCommand( ( tokenTls, tokenTlsVersions,
                                   [ '>>add', tokenTlsVersionsAdd ],
                                   tlsVersionsRule, _enableTlsVersion ) )
SslProfileConfigMode.addCommand( ( tokenTls, tokenTlsVersions,
                                   tokenTlsVersionsRemove,
                                   tlsVersionsRule, _disableTlsVersion ) )

#--------------------------------------------------
# [no|default] fips restrictions
#--------------------------------------------------
def _enableFipsMode( mode ):
   mode.enableFipsMode()

def _disableFipsMode( mode ):
   mode.disableFipsMode()

SslProfileConfigMode.addCommand( ( tokenFips, tokenFipsRestrictions,
                                   _enableFipsMode ) )
SslProfileConfigMode.addCommand( ( BasicCli.noOrDefault, tokenFips,
                                   tokenFipsRestrictions,
                                   _disableFipsMode ) )

#--------------------------------------------------
# [no|default] cipher-list
#--------------------------------------------------
def enableCiphers( mode, ciphers ):
   mode.enableCiphers( ciphers )

def disableCiphers( mode ):
   mode.disableCiphers()

cipherListRule = CliParser.PatternRule( r'[\w:@!+-]+',
                                        helpname='WORD',
                                        helpdesc='Cipher list',
                                        name='ciphers' )

SslProfileConfigMode.addCommand( ( tokenCipherList,
                                   cipherListRule,
                                   enableCiphers ) )

SslProfileConfigMode.addCommand( ( BasicCli.noOrDefault,
                                   tokenCipherList,
                                   BasicCli.trailingGarbage,
                                   disableCiphers ) )

#--------------------------------------------------
# [no|default] certificate command
#--------------------------------------------------
def _setCertKeyPair( mode, certificate, key ):
   mode.setCertKey( certificate, key )

def _noCertificate( mode ):
   mode.noCertKey()

certificateCmd = ( tokenCertificate, certificateNameRule, tokenKey,
                       keyNameRule, _setCertKeyPair )
noCertificateCmd = ( BasicCli.noOrDefault, tokenCertificate,
                         _noCertificate )
SslProfileConfigMode.addCommand( certificateCmd )
SslProfileConfigMode.addCommand( noCertificateCmd )

#--------------------------------------------------
# [no|default] trust certificate commannd
#--------------------------------------------------
def _setTrustedCertificate( mode, certificate ):
   mode.addTrustedCert( certificate )

def _noTrustedCertificate( mode, certificate ):
   mode.noTrustedCert( certificate )

trustedCertNameRule = CliParser.DynamicNameRule( 
               lambda mode: config.profileConfig[ mode.profileName ].trustedCert,
               'Certificate name', 
               name='certificate',
               pattern=r'[A-Za-z0-9_:{}\.\[\]-]+' )

trustCertificateCmd = ( tokenTrust, tokenTrustCertificate, certificateNameRule,
                        _setTrustedCertificate )
noTrustCertificateCmd = ( BasicCli.noOrDefault, tokenTrust,
                          tokenTrustCertificate, trustedCertNameRule,
                          _noTrustedCertificate )
SslProfileConfigMode.addCommand( trustCertificateCmd )
SslProfileConfigMode.addCommand( noTrustCertificateCmd )

#--------------------------------------------------
# [no|default] chain certificate command
#--------------------------------------------------
def _setChainedCertificate( mode, certificate ):
   mode.addChainedCert( certificate )

def _noChainedCertificate( mode, certificate ):
   mode.noChainedCert( certificate )

chainedCertNameRule = CliParser.DynamicNameRule( 
               lambda mode: config.profileConfig[ mode.profileName ].chainedCert,
               'Certificate name', 
               name='certificate',
               pattern=r'[A-Za-z0-9_:{}\.\[\]-]+' )

chainCertificateCmd = ( tokenChain, tokenChainCertificate, certificateNameRule,
                        _setChainedCertificate )
noChainCertificateCmd = ( BasicCli.noOrDefault, tokenChain,
                          tokenChainCertificate, chainedCertNameRule,
                          _noChainedCertificate )
SslProfileConfigMode.addCommand( chainCertificateCmd )
SslProfileConfigMode.addCommand( noChainCertificateCmd )

#--------------------------------------------------
# [no|default] chain certificate requirement include root-ca
#--------------------------------------------------
def _setVerifyChainHasRootCA( mode ):
   mode.setVerifyChainHasRootCA( True )

def _noVerifyChainHasRootCA( mode ):
   mode.setVerifyChainHasRootCA( False )

chainIncludeRootCACmd = ( tokenChain, tokenChainCertificate,
                          tokenRequirement, tokenInclude, tokenRootCA,
                          _setVerifyChainHasRootCA )
noChainIncludeRootCACmd = ( BasicCli.noOrDefault,
                            tokenChain, tokenChainCertificate,
                            tokenRequirement, tokenInclude, tokenRootCA,
                            BasicCli.trailingGarbage,
                            _noVerifyChainHasRootCA )
SslProfileConfigMode.addCommand( chainIncludeRootCACmd )
SslProfileConfigMode.addCommand( noChainIncludeRootCACmd )

#--------------------------------------------------
# [no|default] crl command
#--------------------------------------------------
def _setCrl( mode, certificate ):
   mode.addCrl( certificate )

def _noCrl( mode, crl ):
   mode.noCrl( crl )

crlNameRule = CliParser.DynamicNameRule( 
               lambda mode: config.profileConfig[ mode.profileName ].crl,
               'CRL name', 
               name='crl',
               pattern=r'[A-Za-z0-9_:{}\.\[\]-]+' )

crlCmd = ( tokenCrl, certificateNameRule, _setCrl )
noCrlCmd = ( BasicCli.noOrDefault, tokenCrl, crlNameRule, _noCrl )
SslProfileConfigMode.addCommand( crlCmd )
SslProfileConfigMode.addCommand( noCrlCmd )

#-------------------------------------------------------------------------------
# The "show management security [ssl] certificate [<name>]" command
#-------------------------------------------------------------------------------
def _showSslCert( mode, certName=None ):
   ret = SslModel.Certificates()
   if certName:
      certs = [ certName ]
   else:
      certs = _listdir( Constants.certsDirPath() )
   
   for cert in certs:
      certFile = Constants.certPath( cert )
      if ( os.path.isfile( certFile ) and
           SslCertKey.hasCertificate( open( certFile ).read() ) ):
         ret.certificates[ cert ] = SslCliLib.getCertificateModel( certFile )
   return ret

BasicCli.registerShowCommand( ConfigMgmtMode.tokenManagementForShow,
                              Security.tokenShowSecurity,
                              tokenShowSsl,
                              tokenShowCertificate,
                              [ '>>certName', certificateNameRule ],
                              _showSslCert,
                              privileged=True,
                              cliModel=SslModel.Certificates )

#-------------------------------------------------------------------------------
# The "show management security [ssl] key [<name>]" command
#-------------------------------------------------------------------------------
def _showSslKey( mode, keyName=None ):
   ret = SslModel.PublicKeys()
   if keyName:
      keys = [ keyName ]
   else:
      keys = _listdir( Constants.keysDirPath() )
   
   for key in keys:
      keyFile = Constants.keyPath( key )
      if os.path.isfile( keyFile ):
         ret.publicKeys[ key ] = SslCliLib.getPublicKeyModel( keyFile )
   return ret

BasicCli.registerShowCommand( ConfigMgmtMode.tokenManagementForShow,
                              Security.tokenShowSecurity,
                              tokenShowSsl,
                              tokenShowKey,
                              [ '>>keyName', keyNameRule ],
                              _showSslKey,
                              privileged=True,
                              cliModel=SslModel.PublicKeys )

#-------------------------------------------------------------------------------
# The "show management security [ssl] crl [<name>]" command
#-------------------------------------------------------------------------------
def _showSslCrl( mode, crlName=None ):
   ret = SslModel.Crls()
   if crlName:
      crls = [ crlName ]
   else:
      crls = _listdir( Constants.certsDirPath() )

   for crl in crls:
      crlFile = Constants.certPath( crl )
      if ( os.path.isfile( crlFile ) and
           SslCertKey.hasCrl( open( crlFile ).read() ) ):
         ret.crls[ crl ] = SslCliLib.getCrlModel( crlFile )
   return ret

BasicCli.registerShowCommand( ConfigMgmtMode.tokenManagementForShow,
                              Security.tokenShowSecurity,
                              tokenShowSsl,
                              tokenShowCrl,
                              [ '>>crlName', certificateNameRule ],
                              _showSslCrl,
                              privileged=True,
                              cliModel=SslModel.Crls )

#-------------------------------------------------------------------------------
# The "show management security ssl diffie-hellman" command
#-------------------------------------------------------------------------------
def _dhparamsResetAttempted():
   return status.dhparamsResetProcessed or status.dhparamsLastResetFailed 

def _showSslDhparams( mode ):
   ret = SslModel.DiffieHellman()
   if status.dhparamsResetInProgress:
      ret.dhparamsResetInProgress = True
   else:
      if _dhparamsResetAttempted():
         ret.dhparamsLastResetFailed = status.dhparamsLastResetFailed
         if status.dhparamsResetProcessed:
            ret.dhparamsLastSuccessfulReset = int( status.dhparamsResetProcessed + 
                                                   Tac.utcNow() - 
                                                   Tac.now() )
      ret.dhparamsResetInProgress = False
      dhparamsFile = Constants.dhParamPath()
      if os.path.isfile( dhparamsFile ):
         ret.diffieHellmanParameters = SslCliLib.getDhparamsModel( dhparamsFile )
   return ret


BasicCli.registerShowCommand( ConfigMgmtMode.tokenManagementForShow,
                              Security.tokenShowSecurity,
                              tokenShowSsl,
                              tokenShowDh,
                              _showSslDhparams,
                              privileged=True,
                              cliModel=SslModel.DiffieHellman )

#-------------------------------------------------------------------------------
# The "show management security [ssl] profile" command
#-------------------------------------------------------------------------------
def _showSslProfileStatus( mode, profileName=None ):
   ret = SslModel.SslStatus()
   if profileName:
      profiles = [ profileName ]
   else:
      profiles = status.profileStatus
   
   for name in profiles:
      profile = status.profileStatus.get( name )
      if profile is not None:
         profileStatusModel = SslModel.ProfileStatus()
         profileStatusModel.profileState = profile.state
         for err in profile.error.values():
            profErr = SslModel.ProfileError( 
                          errorAttr=err.errorAttr,
                          errorAttrValue=err.errorAttrValue, 
                          errorType=err.errorType )
            profileStatusModel.profileError.append( profErr )
            ret._hasError = True
         ret.profileStatus[ name ] =  profileStatusModel
   return ret
   
   
BasicCli.registerShowCommand( ConfigMgmtMode.tokenManagementForShow,
                              Security.tokenShowSecurity,
                              tokenShowSsl,
                              tokenShowProfile,
                              [ '>>profileName', statusProfileNameRule ],
                              _showSslProfileStatus,
                              cliModel=SslModel.SslStatus )

#-------------------------------------------------------------------------------
# The "reset ssl diffie-hellman parameters" command
#-------------------------------------------------------------------------------
def _doResetDhparams( mode ):
   execRequest.dhparamsResetRequest = Tac.now()

BasicCli.EnableMode.addCommand( ( CliReset.tokenReset,
                                  tokenSsl,
                                  tokenDh,
                                  tokenParameters,
                                  _doResetDhparams ) )

def Plugin( entityManager ):
   global config, status, execRequest
   config = ConfigMount.mount( entityManager, "mgmt/security/ssl/config",
                               "Mgmt::Security::Ssl::Config", "w" )
   status = LazyMount.mount( entityManager, "mgmt/security/ssl/status",
                             "Mgmt::Security::Ssl::Status", "r" )
   execRequest = LazyMount.mount( entityManager, "mgmt/security/ssl/execRequest",
                                  "Mgmt::Security::Ssl::ExecRequest", "w" )
